program OpenGL33_Tutorial1;

// Erstes Tutorial für OpenGL 3.3+
// Erstellt 2014 für www.delphigl.com von hobbyfrickler
// Basiert auf einem Sample der SDL2-Header für Pascal von end.
// Die SDL2-Header sind auf www.delphigl.com erhältlich: http://www.delphigl.com/forum/viewtopic.php?f=13&t=10752
// Die benötigte "dglOpenGL.pas" gibt es hier: http://wiki.delphigl.com/index.php/dglOpenGL.pas


// Es wird ein zweidimensionales Dreieck mit Farbverlauf gezeichnet.
// Benutzte Techniken: VBO, VAO, simpler Vertex- & Fragmentshader,
// Fensterverwaltung und Ereignisbehandlung mittels SDL2.


{$APPTYPE CONSOLE}

// Damit FreePascal mit dem Adreßoperator (@) von Delphi klarkommt schalten
// wir in den Delphi-Mode von FreePascal
{$IFDEF FPC}
  {$MODE DELPHI}
{$ENDIF}

uses
  Classes, SysUtils,
  // Um die externen Bibliotheken OpenGL bzw. SDL2 benutzen zu können müssen
  // dem Compiler die Pfade zur dglOpenGL.pas und sdl2.pas bekannt gemacht werden:
  // Lazarus: Unter Projekt->Projekteinstellungen->Compilereinstellungen->Pfade
  // Delphi: In den allgemeinen Optionen in den Bibliothekspfad eintragen
  dglOpenGL, SDL2;

var
  window: PSDL_Window;                 // Handle auf das SDL-Fenster
  context: TSDL_GLContext;             // Handle auf den OpenGL Kontext
  running: Boolean;                    // Programm läuft solange = true
  // Zwei VBO's, eins enthält die Vertexdaten, das andere die Farbdaten.
  VBO: array[0..1] of GLuint;
  // Ein VAO, in einem VAO werden die VBO's eines Objektes zusammengefasst
  VAO: GLuint;
  programhandle: GLHandle;             // Handle auf Programmobjekt (für Shader)


procedure Setup_SDL2;
begin
  // Initialisieren des SDL-Videosubsystems
  if SDL_Init(SDL_INIT_VIDEO) <> 0 then
  begin
    raise Exception.Create('Initializieren des SDL-Videosubsystems fehlgeschlagen: ' + SDL_GetError);
  end;
  // Ein Fenster in der Mitte des Desktops erstellen, Fenstergröße ist
  // 640x480 und veränderbar, OpenGL-Unterstützung ist aktiviert
  window:= SDL_CreateWindow('OpenGL 3.3 Tutorial 1', SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, 640, 480, SDL_WINDOW_OPENGL or SDL_WINDOW_RESIZABLE);
//  window:= SDL_CreateWindow('OpenGL 3.3 Tutorial 1', 2500, 500, 640, 480, SDL_WINDOW_OPENGL or SDL_WINDOW_RESIZABLE);
  if window = nil then                // Fenster korrekt erstellt?
  begin
    raise Exception.Create('SDL-Fenster konnte nicht erstellt werden: ' + SDL_GetError);
  end;
  // Die gewünschte Minimum OpenGL-Version festlegen
  SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 3);
  SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 3);
  context := SDL_GL_CreateContext(window);    // OpenGL-Kontext erstellen
  if context = nil then               // Kontext korrekt erstellt?
  begin
    raise Exception.Create('OpenGL-Kontext konnte nicht erstellt werden: ' + SDL_GetError);
  end;
end;


procedure Setup_OpenGL;
begin
  // Initialisieren von OpenGL mit den Funktionen aus der dglOpenGL.pas
  if not InitOpenGL then
  begin
    raise Exception.Create('OpenGL konnte nicht initialisiert werden!');
  end;
  ReadExtensions;
  ReadImplementationProperties;
end;


procedure SetViewport_OpenGL(w, h: Integer);
begin
  // Wird beim Ändern der Fenstergröße aufgerufen, passt den OpenGL-Viewport
  // an die neue Fenstergröße an.
  glViewport( 0, 0, w, h );
  // Anmerkung: Nach Änderung der Fenstergröße sehen die Objekte evtl. perspektivisch
  // verzerrt aus, das liegt daran das wir noch keine Projektionsmatrix benutzen.
end;


procedure HandleEvents_SDL;
var
  event: TSDL_Event;
begin
  while SDL_PollEvent(@event) > 0 do
  begin // SDL-Ereignis vorhanden
    case event.type_ of
      SDL_WINDOWEVENT:
      begin // Ereignis betrifft SDL-Fenster
        case event.window.event of
          SDL_WINDOWEVENT_SIZE_CHANGED:
          begin // Fenstergröße wurde verändert
            // Falls mehrere Fenster mit SDL erzeugt wurden testen ob es das
            // "OpenGL-Fenster" ist
            if event.window.windowID = SDL_GetWindowID(window) then
            begin //OpenGL-Viewport anpassen
              // In event.window.data1 & data2 stehen verschiedene Daten, abhängig
              // von der Art des Ereignisses, hier:
              // event.window.data1 = neue Fensterbreite, data2 = neue Fensterhöhe
              SetViewport_OpenGL(event.window.data1, event.window.data2);
            end;
          end;
          SDL_WINDOWEVENT_CLOSE:
          begin // Fenster soll geschlossen werden, z.B. Klick auf Fensterschließsymbol
            if event.window.windowID = SDL_GetWindowID(window) then
            begin
              running:= false; // Hauptschleife soll beendet werden
            end;
          end;
        end;
      end;
    end;
  end;
end;


// Funktion um Fehler von OpenGL-Objekten auszulesen, momentan können Kompilier-
// und Linkerfehler von Shadern ausgelesen werden.
function GetInfoLogMessage(glObject: GLHandle): string;
var
  infolog_length,    // Länge der Meldung im InfoLog (mit Null-Terminierung)
  stringlen: GLInt;  // Länge des ausgelesenen Strings (ohne Null-Terminierung)
begin
  Result:= '';
  if glIsProgram(glObject) then
  begin
    // Länge der Meldung im InfoLog auslesen
    glGetProgramiv(glObject, GL_INFO_LOG_LENGTH, @infolog_length);
    // Größe des Ergebnisstrings anpassen
    SetLength(Result, infolog_length);
    // Meldung in Ergebnisstring kopieren
    glGetProgramInfoLog(glObject, infolog_length, stringlen, @Result[1]);
  end
  else if glIsShader(glObject) then
  begin
    glGetShaderiv(glObject, GL_INFO_LOG_LENGTH, @infolog_length);
    SetLength(Result, infolog_length);
    glGetShaderInfoLog(glObject, infolog_length, stringlen, @Result[1]);
  end;
end;


//Funktion um die Shader zu laden, gibt ein fertiges Programobjekt zurück
function LoadShader(VFilename: string; FFilename: string): GLHandle;
 var
  ProgramObject, VShaderObject, FShaderObject: GLHandle;
  VStrList, FStrList: TStringList;
  VShaderText, FShaderText: AnsiString;
  VShaderLength, FShaderLength, ShaderCount: Integer;
  CompileStatus, LinkStatus: GLint;
begin
  Result:= -1;
  if not FileExists(VFilename) then
  begin
    raise Exception.Create('Shaderdatei nicht gefunden: ' + VFilename);
  end;
  if not FileExists(FFilename) then
  begin
    raise Exception.Create('Shaderdatei nicht gefunden: ' + FFilename);
  end;
  CompileStatus:= 0;
  LinkStatus:= 0;

  VStrList:= TStringList.Create;
  VStrList.LoadFromFile(VFilename);          // Shaderdatei in Stringliste laden
  VShaderText:= VStrList.Text;               // Aus Stringliste in ANSISTRING
  VShaderLength:= Length(VShaderText);       // Länge des Shaders ermitteln
  VStrList.Free;

  FStrList:= TStringList.Create;
  FStrList.LoadFromFile(FFilename);
  FShaderText:= FStrList.Text;
  FShaderLength:= Length(FShaderText);
  FStrList.Free;

  ProgramObject:= glCreateProgram;           // Programobjekt für Shader erstellen

  VShaderObject:= glCreateShader(GL_VERTEX_SHADER);  // Shaderobjekte erstellen
  FShaderObject:= glCreateShader(GL_FRAGMENT_SHADER);
  // Den Shaderobjekten den Quellcode der Shaderdateien zuweisen
  glShaderSource(VShaderObject, 1, @VShaderText, @VShaderLength);
  glShaderSource(FShaderObject, 1, @FShaderText, @FShaderLength);
  // Shader kompilieren
  glCompileShader(VShaderObject);
  glCompileShader(FShaderObject);
  // Kompilierstatus prüfen
  glGetShaderiv(VShaderObject, GL_COMPILE_STATUS, @CompileStatus);
  if (CompileStatus = GL_FALSE) then
  begin
    raise Exception.Create('Fehler beim kompilieren des VertexShader: ' +
                                            GetInfoLogMessage(VShaderObject));
  end;
  glGetShaderiv(FShaderObject, GL_COMPILE_STATUS, @CompileStatus);
  if (CompileStatus = GL_FALSE) then
  begin
   raise Exception.Create('Fehler beim kompilieren des FragmentShader: ' +
                                           GetInfoLogMessage(FShaderObject));
  end;
  // Die Shaderobjekte an das erstellte Programobjekt binden
  glAttachShader(ProgramObject, VShaderObject);
  glAttachShader(ProgramObject, FShaderObject);
  // Überprüfen ob Shader gebunden werden konnten
  glGetProgramiv(ProgramObject, GL_ATTACHED_SHADERS, @ShaderCount);
  if ShaderCount <> 2 then              // Es sollten 2 Shader gebunden sein
  begin
    raise Exception.Create('Fehler beim binden der Shader, gebundene Shader: '
                                                     + IntToStr(ShaderCount));
  end;

  glDeleteShader(VShaderObject);       // Shaderobjekte werden nicht mehr benötigt,
  glDeleteShader(FShaderObject);       // sind jetzt über Programmobjekt verfügbar

  // Programmobjekt linken
  glLinkProgram(ProgramObject);
  // Linkerstatus prüfen
  glGetProgramiv(ProgramObject, GL_LINK_STATUS, @LinkStatus);
  if (LinkStatus = GL_FALSE) then
  begin
    raise Exception.Create('Fehler beim linken der Shader: ' +
                                            GetInfoLogMessage(ProgramObject));
  end;
  // Programmobjekt aktivieren (Shader benutzen)
  glUseProgram(ProgramObject);
  Result:= ProgramObject;
end;


procedure CreateScene;
const
  // Array mit Vertexpositionsdaten für das Dreieck (3 Vertexpunkte mit jeweils 3 GLFloat)
  fDreieck: array[0..8] of GLfloat =  (-0.7, -0.5, -1.0, 0.7, -0.5, -1.0, 0.0, 0.7, -1.0);
  // Array mit Farbdaten für Vertexe für das Dreieck (3 Vertexpunkte mit jeweils 3 GLFloat)
  fDreieckFarben: array[0..8] of GLfloat = (1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.0);
begin
  // Das VAO erstellen
  glGenVertexArrays(1, @VAO);
  // Zwei VBO's erstellen
  glGenBuffers(2, @VBO);

  // Aktivieren des VAO
  glBindVertexArray(VAO);
  // Aktivieren des ersten VBO im VAO = Positionsdaten des Dreiecks
  glBindBuffer(GL_ARRAY_BUFFER, VBO[0]);
  // VBO mit Daten aus Array befüllen und an GPU schicken
  glBufferData(GL_ARRAY_BUFFER, Length(fDreieck) * SizeOf(GLfloat), @fDreieck, GL_STATIC_DRAW);
  // Wir wollen das erste VBO "erklären"
  glEnableVertexAttribArray(0);
  // Hier teilt man OpenGL mit wie die Daten im VBO zu interpretieren sind
  // 0 = Index des Attributs,
  // 3 = Anzahl der Elemente pro Vertex (x,y,z),
  // GL_FLOAT = Datentyp der Daten,
  // FALSE = Kein normalisieren ausführen (Daten werden nicht verändert),
  // 12 = Byte Offset zwischen 2 Attributen (3 * SizeOf(GLfloat)), in diesem Beispiel könnte man auch 0 angeben (weil alle Elemente direkt aufeinanderfolgen)
  // Pointer(0) = Zeiger auf erstes zu benutzendes Element im Buffer
  glVertexAttribPointer(0, 3, GL_FLOAT, FALSE, 12, Pointer(0));
  // Aktivieren des zweiten VBO im VAO = Farbdaten Dreieck
  glBindBuffer(GL_ARRAY_BUFFER, VBO[1]);
  glBufferData(GL_ARRAY_BUFFER, Length(fDreieckFarben) * SizeOf(GLfloat), @fDreieckFarben, GL_STATIC_DRAW);
  glEnableVertexAttribArray(1);         // Wir wollen das zweite VBO "erklären"
  glVertexAttribPointer(1, 3, GL_FLOAT, FALSE, 12, Pointer(0));
end;


procedure RenderScene;
begin
  // Letztes Bild und Tiefenpuffer löschen
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
  // Aktivieren des VAO
  glBindVertexArray(VAO);
  // Zeichnen des Dreiecks mittels VBO's (im VAO), GL_TRIANGLES bewirkt das Dreiecke erstellt
  // werden, 0 = Startindex im VBO, 3 = Anzahl der verwendeten Vertexpunkte für das Objekt
  glDrawArrays(GL_TRIANGLES, 0, 3);
  // Framebuffer tauschen
  SDL_GL_SwapWindow(window);
end;


begin
  try
    Setup_SDL2();                     // Fenster und Kontext erstellen
    Setup_OpenGL();                   // OpenGL initialisieren
    SetViewport_OpenGL(640, 480);     // Viewport setzen
    programhandle:= LoadShader(       // Shader laden
                               'shader' + PathDelim + 'VertexShader.txt',
                               'shader' + PathDelim + 'FragmentShader.txt');
    CreateScene();                    // Szene erstellen

    running:= true;
    while running and (window <> nil) do
    begin                // Hauptschleife für Rendern und Ereignisbehandlung
      HandleEvents_SDL();
      RenderScene();
    end;

    // OpenGL Resourcen freigeben
    glDeleteBuffers(2, @VBO);
    glDeleteVertexArrays(1, @VAO);
    if programhandle <> -1 then glDeleteProgram(programhandle);
    // SDL Resourcen freigeben
    SDL_GL_DeleteContext(context);
    SDL_DestroyWindow(window);
    window:= nil;
    SDL_Quit();
  except
    on E: Exception do
      SDL_ShowSimpleMessageBox(SDL_MESSAGEBOX_ERROR, PAnsiChar('Fehler'),
                                                 PAnsiChar(E.Message), window);
  end;
end.

